﻿using Abt.Controls.SciChart;
using Abt.Controls.SciChart.ChartModifiers;
using Abt.Controls.SciChart.Visuals;

using System;
using System.Diagnostics;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;

namespace ExampleProject
{
  public static class FrameworkElementExtensions
  {
    public static T FindLogicalParent<T>(this FrameworkElement element) where T : FrameworkElement
    {
      var parent = (FrameworkElement)element.Parent;

      if (parent is T)
        return (T)parent;

      return parent.FindLogicalParent<T>();
    }
  }

  public class ChartCursorModifier : ChartModifierBase
  {
    #region Properties

    public static readonly DependencyProperty GeometryProperty =
     DependencyProperty.Register("Geometry", typeof(TestList), typeof(ChartCursorModifier),
     new PropertyMetadata(default(TestList)));

    public static readonly DependencyProperty SelectedChartLineProperty =
    DependencyProperty.Register("SelectedChartLine", typeof(int), typeof(ChartCursorModifier),
    new PropertyMetadata(default(int)));

    public static readonly DependencyProperty CurrentPositionProperty =
     DependencyProperty.Register("CurrentPosition", typeof(double), typeof(ChartCursorModifier),
     new PropertyMetadata(default(double)));

    public static readonly DependencyProperty CurrentElementProperty =
     DependencyProperty.Register("CurrentElement", typeof(TestElement), typeof(ChartCursorModifier),
     new PropertyMetadata(default(TestElement)));

    public static readonly DependencyProperty ChartXRangeProperty =
     DependencyProperty.Register("ChartXRange", typeof(DoubleRange), typeof(ChartCursorModifier),
     new PropertyMetadata(default(DoubleRange)));

    public TestElement CurrentElement
    {
      get
      {
        return (TestElement)GetValue(CurrentElementProperty);
      }
      set
      {
        if (value != CurrentElement)
          SetValue(CurrentElementProperty, value);
      }
    }

    public DoubleRange ChartXRange
    {
      get
      {
        return (DoubleRange)GetValue(ChartXRangeProperty);
      }
      set
      {
        if (value != ChartXRange)
          SetValue(ChartXRangeProperty, value);
      }
    }

    public bool IsOpen;

    public TestList Geometry
    {
      get
      {
        return (TestList)GetValue(GeometryProperty);
      }
      set
      {
        SetValue(GeometryProperty, value);
      }
    }

    public double CurrentPosition
    {
      get
      {
        return (double)GetValue(CurrentPositionProperty);
      }
      set
      {
        SetValue(CurrentPositionProperty, value);
      }
    }

    public int SelectedChartLine
    {
      get
      {
        return (int)GetValue(SelectedChartLineProperty);
      }
      set
      {
        SetValue(SelectedChartLineProperty, value);
      }
    }

    #endregion Properties

    #region private Members

    private double jumpWidth;
    private System.Windows.Shapes.Line _line;
    private Point _CurrentPoint = new Point();
    private bool _MoveCursor = false;

    private double ChartSettlingRange
    {
      get
      {
        return (double)ChartXRange.Diff * 0.07;
      }
    }

    #endregion private Members

    #region Constructor

    /// <summary>
    /// Constructor
    /// </summary>
    public ChartCursorModifier()
      : base()
    {
      this._line = new System.Windows.Shapes.Line() { Stroke = new SolidColorBrush(Colors.DarkBlue), StrokeThickness = 2, IsHitTestVisible = false };
      jumpWidth = 1;
    }

    #endregion Constructor

    #region public Methods

    public override void OnModifierDoubleClick(ModifierMouseArgs e)
    {
      base.OnModifierDoubleClick(e);
      if (e.MouseButtons == MouseButtons.Left)
      {
        var xCalc = this.XAxis.GetCurrentCoordinateCalculator();
        var pt = GetPointRelativeTo(e.MousePoint, this.ModifierSurface);
        _CurrentPoint = pt;
        double xDataValue = xCalc.GetDataValue(pt.X);
      }
    }

    /// <summary>
    /// Event if mousebutton is pressed
    /// </summary>
    /// <param name="e"></param>
    public override void OnModifierMouseDown(ModifierMouseArgs e)
    {
      if (!(Keyboard.Modifiers == ModifierKeys.Alt))
      {
        IsOpen = (e.MouseButtons == MouseButtons.Right) && (e.Modifier == MouseModifier.Alt);

        if (e.MouseButtons == MouseButtons.Left)
        {
          base.OnModifierMouseDown(e);
          _MoveCursor = true;
          var pt = GetPointRelativeTo(e.MousePoint, this.ModifierSurface);
          _CurrentPoint = pt;
          SetCursorToPointByMouse(pt);
        }
      }
      else if (e.MouseButtons == MouseButtons.Left)
      {
      }
    }

    private void SetCursorToPointByMouse(Point pt)
    {
      var allSeries = this.ParentSurface.RenderableSeries;

      var xCalc = this.XAxis.GetCurrentCoordinateCalculator();
      if (xCalc != null)
      {
        double xDataValue = 0;

        xDataValue = xCalc.GetDataValue(pt.X);

        TestElement elem = null;

        if (allSeries[1].DataSeries != null)
        {
          elem = GetElementAtPositionX(xDataValue);
        }

        SetVisibleRangeofChart(xDataValue, 1);
        SetLineToPosition(Math.Round(xDataValue, 0));
      }
    }

    /// <summary>
    /// Event if mousebutton is released
    /// </summary>
    /// <param name="e"></param>
    public override void OnModifierMouseUp(ModifierMouseArgs e)
    {
      if (!(Keyboard.Modifiers == ModifierKeys.Alt))
      {
        // Open menu if right mouse btn was clicked and Alt btn was pressed
        // IsOpen should be OneWay-bound to SciChartSurface.ContextMenu.IsOpen
        IsOpen = false;

        if (e.MouseButtons == MouseButtons.Left)
        {
          base.OnModifierMouseUp(e);
          _MoveCursor = false;
        }
      }
    }

    /// <summary>
    /// Resizes Line after resize of window or Gridsplitter
    /// </summary>
    /// <param name="e"></param>
    public override void OnParentSurfaceResized(Abt.Controls.SciChart.SciChartResizedMessage e)
    {
      base.OnParentSurfaceResized(e);

      var allSeries = this.ParentSurface.RenderableSeries;

      //this.XAxis.OnBeginRenderPass();
      // Translates the mouse point to chart area, e.g. when you have left axis

      var xCalc = this.XAxis.GetCurrentCoordinateCalculator();

      //Point mousePoint = e.GetPosition(this.ModifierSurface as UIElement);
      if (xCalc != null)
      {
        double xDataValue = xCalc.GetDataValue(_line.X1);

        TestElement elem = null;

        if (allSeries[1].DataSeries != null)
        {
          elem = GetElementAtPositionX(xDataValue);
        }
        AddCursorLineToPos(new Point(_line.X1, 0), xDataValue);
      }
    }

    /// <summary>
    /// Event if mouse is moving
    /// </summary>
    /// <param name="e"></param>
    public override void OnModifierMouseMove(ModifierMouseArgs e)
    {
      if (_MoveCursor)
      {
        base.OnModifierMouseMove(e); // Prevent MouseMove from going to base class

        //// Translates the mouse point to chart area, e.g. when you have left axis
        var pt = GetPointRelativeTo(e.MousePoint, this.ModifierSurface);
        _CurrentPoint = pt;
        jumpWidth = 1;
        SetCursorToPointByMouse(pt);
      }
    }

    /// <summary>
    /// Sets Cursorline at the beginning of the Geometryelement
    /// </summary>
    /// <param name="coorDataVM">Element which Cursorline is set to</param>
    public void SetToBeginning(TestElement geomDataVM)
    {
      var allSeries = this.ParentSurface.RenderableSeries;
      try
      {
        this.XAxis.OnBeginRenderPass();
      }
      catch (Exception EX)
      { Debug.WriteLine(EX.Message); }
      var xCalc = this.XAxis.GetCurrentCoordinateCalculator();
      if (xCalc != null)
      {
        double xDataValue = xCalc.GetCoordinate(geomDataVM.StartPosition);
        if (!Double.IsInfinity(xDataValue) && !Double.IsNaN(xDataValue))
        {
          Point pt = new Point(xDataValue, 0);
          AddCursorLineToPos(pt, xCalc.GetDataValue(xDataValue));
        }
      }
    }

    public void SetToBeginningKeys(TestElement coorDataVM, bool WithFullRange = false)
    {
      var allSeries = this.ParentSurface.RenderableSeries;

      // Translates the mouse point to chart area, e.g. when you have left axis
      if (!WithFullRange)
        SetVisibleRangeofChart(coorDataVM);
      else
        SetVisibleRangeofChartWithRange(coorDataVM);
      this.XAxis.OnBeginRenderPass();
      var xCalc = this.XAxis.GetCurrentCoordinateCalculator();

      Point pt = new Point(xCalc.GetCoordinate(coorDataVM.StartPosition), 0);
      AddCursorLineToPos(pt, coorDataVM.StartPosition);

      _CurrentPoint = pt;
    }

    #endregion public Methods

    #region private Methods

    /// <summary>
    /// Set Cursorline to position x
    /// </summary>
    /// <param name="pt">Point to set Cursorline</param>
    private void SetCursorToPoint(Point pt, bool roundTo1m)
    {
      SetLineToPoint(pt, roundTo1m);
    }

    ///// <summary>
    ///// Set Cursorline to position x
    ///// </summary>
    ///// <param name="pt">Point to set Cursorline</param>
    private void AddCursorLineToPos(Point pt, double value)
    {
      double shiftingValue = 0;
      double correctionValue = 0;

      TestElement element = GetElementAtPositionX(value);
      if (element != null)
      {
        correctionValue = element.Value;
        shiftingValue = element.Value;
      }
      var textPosition = new Border()

      {
        IsHitTestVisible = false,
        BorderBrush = new SolidColorBrush(Colors.DarkBlue),
        BorderThickness = new Thickness(1),
        Background = new SolidColorBrush(Colors.DarkBlue),
        CornerRadius = new CornerRadius(2, 2, 2, 2),
        Child = new TextBlock()
        {
          Foreground = new SolidColorBrush(Colors.White),
          Text = string.Format("{0} m", Math.Round(value, 1)),
          FontSize = 13,
          Margin = new Thickness(3),
        }
      };

      var textShifting = new Border()
       {
         IsHitTestVisible = false,
         BorderBrush = new SolidColorBrush(Colors.DarkBlue),
         BorderThickness = new Thickness(1),
         Background = new SolidColorBrush(Colors.DarkBlue),
         CornerRadius = new CornerRadius(2, 2, 2, 2),
         Child = new TextBlock()
         {
           Foreground = new SolidColorBrush(Colors.White),
           Text = string.Format("{0:0.0} mm", Math.Round(shiftingValue, 1)),
           FontSize = 13,
           Margin = new Thickness(3),
         }
       };

      var textCorrection = new Border()
      {
        IsHitTestVisible = false,
        BorderBrush = new SolidColorBrush(Colors.DarkBlue),
        BorderThickness = new Thickness(1),
        Background = new SolidColorBrush(Colors.DarkBlue),
        CornerRadius = new CornerRadius(2, 2, 2, 2),
        Child = new TextBlock()
        {
          Foreground = new SolidColorBrush(Colors.White),
          Text = string.Format("{0:0.0} mm", Math.Round(correctionValue, 1)),
          FontSize = 13,
          Margin = new Thickness(3),
        }
      };
      var elem = GetElementAtPositionX(value);
      CurrentElement = elem;

      CurrentPosition = value;

      _CurrentPoint = pt;
      _line.Y1 = 0;
      _line.Y2 = ModifierSurface.ActualHeight;
      _line.X1 = pt.X;
      _line.X2 = pt.X;

      Canvas.SetLeft(textPosition, pt.X);
      Canvas.SetBottom(textPosition, 0);
      Canvas.SetLeft(textShifting, pt.X);
      Canvas.SetBottom(textShifting, ModifierSurface.ActualHeight - 30);
      Canvas.SetLeft(textCorrection, pt.X);
      Canvas.SetBottom(textCorrection, ModifierSurface.ActualHeight / 2 - 30);

      this.ModifierSurface.Clear();
      this.ModifierSurface.Children.Add(_line);
      this.ModifierSurface.Children.Add(textPosition);
      this.ModifierSurface.Children.Add(textShifting);
      this.ModifierSurface.Children.Add(textCorrection);
      this.ParentSurface.InvalidateElement();
    }

    /// <summary>
    ///CursorLine wird auf aktuelle Position gesetzt ohne das Property CurrentElement updaten
    /// </summary>
    public void SetLineToCurrentPointWithoutSelectingElement()
    {
      if (_CurrentPoint != null)
      {
        var allSeries = this.ParentSurface.RenderableSeries;
        try
        {
          this.XAxis.OnBeginRenderPass();
        }
        catch (Exception Ex)
        {
          Debug.WriteLine(Ex.Message);
        }
        // Translates the mouse point to chart area, e.g. when you have left axis
        var xCalc = this.XAxis.GetCurrentCoordinateCalculator();

        if (xCalc != null)
        {
          double xDataValue = xCalc.GetDataValue(_CurrentPoint.X);
          AddCursorLineToPos(_CurrentPoint, xDataValue);
        }
      }
    }

    /// <summary>
    ///CursorLine wird auf aktuelle Position gesetzt und das Property CurrentElement wird auf das aktuelle Element gesetzt
    /// </summary>
    public void SetLineToCurrentPointWithSelectingElement()
    {
      if (_CurrentPoint != null)
      {
        var allSeries = this.ParentSurface.RenderableSeries;
        try
        {
          this.XAxis.OnBeginRenderPass();
        }
        catch (Exception Ex)
        {
          Debug.WriteLine(Ex.Message);
        }
        // Translates the mouse point to chart area, e.g. when you have left axis
        var xCalc = this.XAxis.GetCurrentCoordinateCalculator();

        if (xCalc != null)
        {
          double xDataValue = xCalc.GetDataValue(_CurrentPoint.X);

          AddCursorLineToPos(_CurrentPoint, xDataValue);
        }
      }
    }

    public override void OnParentSurfaceRendered(SciChartRenderedMessage e)
    {
      base.OnParentSurfaceRendered(e);
      var xCalc = this.XAxis.GetCurrentCoordinateCalculator();
      double xDataValue = xCalc.GetCoordinate(CurrentPosition);

      AddCursorLineToPos(new Point(xDataValue, 0), CurrentPosition);
    }

    private void SetLineToPoint(Point pt, bool roundTo1m)
    {
      var allSeries = this.ParentSurface.RenderableSeries;

      var xCalc = this.XAxis.GetCurrentCoordinateCalculator();
      if (xCalc != null)
      {
        //Point mousePoint = e.GetPosition(this.ModifierSurface as UIElement);
        double xDataValue = xCalc.GetDataValue(pt.X);

        SetVisibleRangeofChart(xDataValue, jumpWidth);
        if (roundTo1m)
          SetLineToPosition(Math.Round(xDataValue, 0));
        else
          SetLineToPosition(xDataValue);
      }
    }

    public void SetLineToPosition(double position)
    {
      var allSeries = this.ParentSurface.RenderableSeries;
      this.XAxis.OnBeginRenderPass();
      // Translates the mouse point to chart area, e.g. when you have left axis

      var xCalc = this.XAxis.GetCurrentCoordinateCalculator();

      //Point mousePoint = e.GetPosition(this.ModifierSurface as UIElement);
      double xDataValue = xCalc.GetCoordinate(position);
      Point pt = new Point(xDataValue, 0);

      AddCursorLineToPos(pt, position);
      _CurrentPoint = pt;
    }

    private TestElement GetLastElement(TestList geometry)
    {
      var list = geometry;
      if (list.Count() > 0)
        return list.Last();
      return null;
    }

    private TestElement GetFirstElement(TestList geometry)
    {
      var list = geometry;
      if (list.Count() > 0)
        return list.First();
      return null;
    }

    /// <summary>
    /// Returns next element of current Element
    /// </summary>
    /// <param name="Geometry">Geometry Direction/Superelevation or Longitudinal</param>
    /// <returns></returns>
    private TestElement GetNextElement(TestList geometry)
    {
      int idx = geometry.IndexOf(CurrentElement);
      if (idx < geometry.Count - 1 && idx != -1)
      {
        return geometry[++idx];
      }
      else
        return CurrentElement;
    }

    /// <summary>
    /// Returns preivious element of current Element
    /// </summary>
    /// <param name="Geometry">Geometry Direction/Superelevation or Longitudinal</param>
    /// <returns></returns>
    private TestElement GetPreviousElement(TestList geometry)
    {
      int idx = geometry.IndexOf(CurrentElement);
      if (idx > 0 && idx != -1)
      {
        return geometry[--idx];
      }
      else
        return CurrentElement;
    }

    /// <summary>
    /// Returns Element of View at Position X
    /// </summary>
    /// <param name="posX"> Current Cursor Position</param>
    /// <param name="elemToCheck"> Type of View</param>
    /// <param name="Geometry">Nominal Geometry</param>
    /// <returns></returns>
    private TestElement GetElementAtPositionX(double posX)
    {
      TestElement elem = null;
      if (Geometry != null)
      {
        elem = Geometry.FirstOrDefault(x => (x.StartPosition >= Convert.ToDouble(posX)));

        return elem;
      }
      return null;
    }

    #endregion private Methods

    #region KeyboardMoves

    public void GeometryCursorModifier_PreviewKeyDown(object sender, KeyEventArgs e)
    {
      double range = (double)this.XAxis.VisibleRange.Diff;
      var xCalc = this.XAxis.GetCurrentCoordinateCalculator();
      double rangeXAxes = (double)xCalc.GetCoordinate((double)this.XAxis.VisibleRange.Max) - xCalc.GetCoordinate((double)this.XAxis.VisibleRange.Min);

      double fracLiRe = (rangeXAxes / range) * 1.0;
      jumpWidth = 1.0;
      if (Keyboard.IsKeyDown(Key.D1))  //dm
      {
        fracLiRe = (rangeXAxes / range) * 0.1;
        jumpWidth = 0.1;
      }
      else if (Keyboard.IsKeyDown(Key.D2)) //10m
      {
        fracLiRe = (rangeXAxes / range) * 10.0;
        jumpWidth = 10;
      }
      else if (Keyboard.IsKeyDown(Key.D3))  //10%
      {
        fracLiRe = (rangeXAxes / range) * range * 0.1;
        jumpWidth = range * 0.1;
      }
      else if (Keyboard.IsKeyDown(Key.PageDown) || Keyboard.IsKeyDown(Key.PageUp))  //10%
      {
        fracLiRe = (rangeXAxes / range) * range;
        jumpWidth = range;
      }

      if (e.Key == Key.Right)
      {
        _CurrentPoint.X += fracLiRe;

        UpdateRollover();
      }
      else if (e.Key == Key.Left)
      {
        _CurrentPoint.X -= fracLiRe;

        UpdateRollover();
      }
      else if (e.Key == Key.PageUp)
      {
        SetPageUp();
        SetLineToCurrentPointWithSelectingElement();
      }
      else if (e.Key == Key.PageDown)
      {
        SetPageDown();
        SetLineToCurrentPointWithSelectingElement();
      }
    }

    public void Move10mUpDown(bool Up)
    {
      double range = (double)this.XAxis.VisibleRange.Diff;
      var xCalc = this.XAxis.GetCurrentCoordinateCalculator();
      double rangeXAxes = (double)xCalc.GetCoordinate((double)this.XAxis.VisibleRange.Max) - xCalc.GetCoordinate((double)this.XAxis.VisibleRange.Min);

      double fracUpDown = (rangeXAxes / range) * 10.0;
      jumpWidth = 10;
      if (Up)
      {
        _CurrentPoint.X += fracUpDown;

        UpdateRollover();
      }
      else
      {
        _CurrentPoint.X -= fracUpDown;

        UpdateRollover();
      }
    }

    public void MoveToNexElement()
    {
      TestElement elem = null;

      elem = GetNextElement(Geometry);

      SetToBeginningKeys(elem);
    }

    public void MoveToPreviousElement()
    {
      TestElement elem = null;

      elem = GetPreviousElement(Geometry);

      SetToBeginningKeys(elem);
    }

    public void MoveToFirstElement()
    {
      TestElement elem = null;

      elem = GetFirstElement(Geometry);

      SetToBeginningKeys(elem);
    }

    public void MoveToLastElement()
    {
      TestElement elem = null;

      elem = GetLastElement(Geometry);

      if (elem != null)
        SetToBeginningKeys(elem, true);
    }

    public override void OnAttached()
    {
      base.OnAttached();
      var scichart = (ParentSurface as SciChartSurface);
      scichart.FindLogicalParent<ChartView>().PreviewKeyDown +=
          new KeyEventHandler(GeometryCursorModifier_PreviewKeyDown);
    }

    public override void OnDetached()
    {
      base.OnDetached();
      var scichart = (ParentSurface as SciChartSurface);
      scichart.FindLogicalParent<ChartView>().PreviewKeyDown -=
          new KeyEventHandler(GeometryCursorModifier_PreviewKeyDown);
    }

    private void UpdateRollover()
    {
      SetCursorToPoint(_CurrentPoint, false);
    }

    #endregion KeyboardMoves

    #region ChartXRange

    /// <summary>
    /// Sets Min Max of XAxis depending of the selected element
    /// </summary>
    /// <param name="EndPosition"> Largest EndPosition of Geometryelement</param>
    /// <param name="geomData"> Current Selected Element</param>
    public void SetVisibleRangeofChart(TestElement geomData)
    {
      double min, max;
      var currentRange = this.XAxis.VisibleRange as DoubleRange;
      if (currentRange != null)
      {
        if (geomData.StartPosition < currentRange.Min || geomData.StartPosition > currentRange.Max - ChartSettlingRange)
        {
          min = geomData.StartPosition - ChartSettlingRange;
          max = geomData.StartPosition + currentRange.Diff - ChartSettlingRange;
          ChartXRange = new DoubleRange(min, max);
        }
      }
    }

    /// <summary>
    /// Sets Min Max of XAxis depending of the selected element
    /// </summary>
    /// <param name="position"> Position of Element(StartPosition)</param>
    public void SetVisibleRangeofChart(double position)
    {
      double min, max;
      var currentRange = this.XAxis.VisibleRange as DoubleRange;
      if (currentRange != null)
      {
        if (position < currentRange.Min || position > currentRange.Max - ChartSettlingRange)
        {
          min = position - ChartSettlingRange;
          max = position + currentRange.Diff - ChartSettlingRange;
          ChartXRange = new DoubleRange(min, max);
        }
      }
    }

    /// <summary>
    /// Sets Min Max of XAxis depending of the current X-Value of X-Axis
    /// </summary>
    /// <param name="currentPosition"></param>
    public void SetVisibleRangeofChart(double currentPosition, double jumpWidth)
    {
      var currentRange = ChartXRange as DoubleRange;
      double min = currentRange.Min, max = currentRange.Max;
      if (currentPosition >= currentRange.Max - ChartSettlingRange)
      {
        min = currentRange.Min + jumpWidth;
        max = currentRange.Max + jumpWidth;
        ChartXRange = new DoubleRange(min, max);
      }
      else if (currentPosition <= currentRange.Min + ChartSettlingRange)
      {
        min = currentRange.Min - jumpWidth;
        max = currentRange.Max - jumpWidth;
        ChartXRange = new DoubleRange(min, max);
      }
      else
      {
        ChartXRange = new DoubleRange(min, max);
      }
    }

    /// <summary>
    /// Sets Min Max of XAxis depending of the selected element
    /// </summary>
    /// <param name="Endposition"> Largest Endposition of Geometryelement</param>
    /// <param name="geomDataDirSup"> Current Selected Element</param>
    public void SetVisibleRangeofChartWithRange(TestElement corrValue)
    {
      double min, max;
      var currentRange = this.XAxis.VisibleRange as DoubleRange;
      if (currentRange != null && corrValue != null)
      {
        if (corrValue.StartPosition < currentRange.Min || corrValue.StartPosition > currentRange.Max - ChartSettlingRange)
        {
          min = corrValue.StartPosition - currentRange.Diff + ChartSettlingRange;
          max = corrValue.StartPosition + ChartSettlingRange;
          ChartXRange = new DoubleRange(min, max);
        }
      }
    }

    public void SetPageUp()
    {
      var currentRange = ChartXRange as DoubleRange;
      var dif = ChartXRange.Diff;
      ChartXRange = new DoubleRange(currentRange.Min + dif, currentRange.Max + dif);
    }

    public void SetPageDown()
    {
      var currentRange = ChartXRange as DoubleRange;
      var dif = ChartXRange.Diff;
      ChartXRange = new DoubleRange(currentRange.Min - dif, currentRange.Max - dif);
    }

    #endregion ChartXRange
  }
}